package PureMessage::Policy::CreditCard;

use strict;
use warnings;

use HTML::Parser;

use base 'PureMessage::Policy::Plugin';
use PureMessage::Policy::Words qw(message_words);
use PureMessage::Policy::CreditCard::Config qw(%config);
use PureMessage::Policy::Plugin::Utils qw(
        expand_localized_tmpl
        replace_nested_part
        set_attachments 
        );
use PureMessage::Policy::MessageText qw(match_attachment_text
                                        per_attachment_byte_limit_exceeded
                                        per_message_time_limit_exceeded
                                        current_attachment_part_id
        );

our @EXPORT_OK = qw(pmx_credit_card);
our $VERSION   = '6.403';

my %commands = (
    pmx_credit_card                   => \&pmx_credit_card,
);

sub new {
    my $class = shift;
    my %opts = @_;
    return bless \%opts, $class;
}

sub commands {
    my $self = shift;
    return %commands;
}

sub pmx_credit_card {
    my ($msg,$args) = @_;

    my $no_early_out = $args->tag('scanall') ? 1 : 0;

    # Clear the cantscan hash attribute
    #   Before the release of PMX v6.4.3, there used to be a configurable parameter - 'creditcard_number_limit" in /opt/pmx6/etc/creditcard.conf 
    #   which determined how many numeric sections PMX will scan in a given message to find a CC number. If this limit exceeds, PMX used to
    #   set the cantscan_map. This behavior is removed now. However the policy.siv may still contain pmx_cantscan following pmx_credit_card.
    $msg->set_attr('cantscan_map', [{'type' => 'cc', 'ids' => []}], 1);
    my $cantscan_map = { type => 'cc', ids => [] };

    my $overlap_buf = '';
    my $cc_found = 0;
    my $html;
    $html = HTML::Parser->new(
	api_version => 3,
	handlers => {
	    text => [sub {
            if(_credit_card_check($overlap_buf . $_[0])) {
                $cc_found = 1;
                # from 'perldoc HTML::Parser': Calling $p->eof inside a handler will terminate parsing at that point 
                #   and cause $p->parse to return a FALSE value.
                $overlap_buf = '';
                $html->eof;
            }
            $overlap_buf = substr($_[0], -30);
	    }, 'dtext'],
	},
    );

    if ($msg->header_exists('Subject')) {
	    for my $h ($msg->header_decoded('Subject')) {
	        last unless ($html->parse($h));
            # from 'perldoc HTML::Parser': Calling the $p->eof method outside a handler callback will flush any 
            #   remaining buffered text (which triggers the "text" event if there is any remaining text).
	        $html->eof;
            $overlap_buf = '';
	    }
    }

    if ($cc_found) {
        $msg->log("DEBUG", "pmx_credit_card: credit card found");
        return 1;
    }

    my @parts = $msg;
    while (my $part = shift @parts) {
    	my @ids = $part->parts;
    	unshift @parts, map { $part->part($_) } @ids if @ids;
    	my $ct = $part->content_type;
    	if ($ct =~ m#^text/(?:plain|html)#i) {
    	    my $offset = 0;
    	    while (my $len = $part->body(my $buf, $offset, 8192)) {
        		last unless ($html->parse($buf));
        		$offset += $len;
    	    }
    	    $html->eof;
            $overlap_buf = '';
            if ($cc_found) {
                $msg->log("DEBUG", "pmx_credit_card: credit card found");
                return 1;
            }
    	}
    }

    if ($args->tag('attachments')) {
        my @matching_parts;
        my @non_matching_parts;

        my @matches = match_attachment_text($msg,
                            sub {
                                my $result = _credit_card_check($_[0]);

                                # Returns an array to make match_attachment_text() happy.
                                # (match_attachment_text() is expecting an array of length 2 to be returned
                                # from the matcher callback) 
                                # We never want to return an actual credit card number for obvious privacy
                                # concerns.
                                return ($result, '');
                            },
                matching_parts     => \@matching_parts,
                non_matching_parts => \@non_matching_parts,
                no_early_out       => $no_early_out,
            );

        my $modified_cantscan = $msg->get_attr('cantscan_map');
        if (defined $modified_cantscan and my $last_cantscan = $modified_cantscan->[-1]) {
            push @{$last_cantscan->{ids}}, @{$cantscan_map->{ids}};
            $msg->set_attr('cantscan_map', $modified_cantscan, 1);

        } else {
            $msg->set_attr('cantscan_map', [$cantscan_map], 1);
        }

        if (@matches) {
            $msg->log("DEBUG", "pmx_credit_card :attachments - credit card found");
            $msg->template("MATCHED_FILE", $matches[0]->{file});
            set_attachments($msg, \@matching_parts, \@non_matching_parts);

            return 1;
        }

        set_attachments($msg, \@matching_parts, \@non_matching_parts);

        return (pmx_credit_card_limit_exceeded($msg)
            || $msg->{per_message_savi_cantscan_errors});
    }

    return 0;
}

sub pmx_credit_card_limit_exceeded {
    my $msg = shift;
    return $msg->get_attr('cc_limit_exceeded')
        || per_attachment_byte_limit_exceeded($msg)
        || per_message_time_limit_exceeded($msg);
}

sub _credit_card_check {

    # Do NOT change the value of $str (we need something like http://search.cpan.org/~sanko/Readonly-2.05/lib/Readonly.pm)
    # See 'perldoc HTML::Parser': Handlers invoked should not attempt modify the $string in-place until $p->parse returns.
 	my $str = \$_[0];

    # defining space makes the last space in a statement visible
    my $ccr_sp = qr/\ /x;
    
    # The CC number should start either with a boundary or, in case it comes from a PDF form, with an _
    my $ccr_prefix = qr/(?:\b|_)/; 
    
    # To make sure that the number is not a part of a larger numberic sequence or 
    # is not a part of a mathematical sequence, check that it is not preceded by 
    # either of these three character fixed-width patterns (used as a negative lookbehind 
    # inside main regex)
    my $ccr_csb =  # Confirm Start Boundary
    qr/(?:..[+*\/=<>]      # a common mathematical symbol
         |.[+*\/=<>]${ccr_sp}
         |.[^:a-zA-Z]-     # a minus sign except it is like 'No:-' or 'No-'
         |[^:a-zA-Z]-${ccr_sp}
         |[\d )]x${ccr_sp}     # an 'x' as a math symbol like 5 x 34
         |.\d[. ]          # a fraction like 5.6, or a digit followed by space
       )
    /sx; # The use of 's' flag (dot-matches-all) is very important here. As all flags that you have not set explicitly the qr// operator unsets them using (?-<flag_that_you_have_not_set_explicitly>:).
         # Without the 's' flag, $ccr_csb will byte you when interpolated inside $ccr_main.
         # Ref: http://www.perlmonks.org/bare/?node_id=282640
    
    
    my $ccr_suffix = 
    qr/(?:(?=\D{4})  # the CC number should be followed by four non-digit chars. So "3123123456789012e-12" is not valid.
         |(?:$|_|\b(?!\.\d              # or end of document or an underscore or a word boundary not followed by a decimal point and a digit "3123123456789012.2"
                     |${ccr_sp}?[+=*\/<>-]  #                                                          a math symbol: "3123-1234-1234-1234=-579"
                     |${ccr_sp}x\b          #                                                          a space and 'x' which can mimic a math symbol "3123123456789012 x 3"
                     |${ccr_sp}\d           #                                                          a space and another decimal "3123123456789012 3123123456789012"
                   )
          )
    )
    /x;
    
    my $ccr_main = 
    qr/(?:
            ${ccr_prefix}
            (        # capture group-1: for luhn validation
               (?:   # all possible credit card formats ORed within this non-capturing group
                 # standard 16 digit banking format with standard four groupings
                 [3-6]\d{3}(?:\d{12}(?s:(?<!${ccr_csb}.{16}))  # 3123123456789012
                             |(?:(?:${ccr_sp}\d{4}){3}         # 3123 1234 1234 1234
                                |(?:-\d{4}){3})                # 3123-1234-1234-1234
                              (?s:(?<!${ccr_csb}.{19}))
                           )
    
                 # Roayl Bank of Canada CA - client card, ATM or INTERAC
                 # prefixed 16 digit with flexible groupings, separator can be either space or dash
                 |4519(?s:(?<!${ccr_csb}.{4}))(?:(?:${ccr_sp}?\d){12}  # 4519 123 123 123 123, 4519 1234 1234 1234, 451912 123456 1234, etc
                                                |(?:-?\d){12})         # 4519123-123-123-123,  451912-12-1234-1234, etc
    
                 # 19 digits that should not be prefixed with LinkedIn invitation string, so far Maestro and CUP
                 |6\d{18}(?s:(?<!${ccr_csb}.{19}))(?<![&=]invitationId=.{19}) # 6123456789012345678
    
                 # Maestro 16 to 19 digits
                 |(?:0604|50[13]8|5020|5893|6390)(?s:(?<!${ccr_csb}.{4}))(?:(?:${ccr_sp}?\d){12,15} # 5018 1234 1234 12345
                                                                           |(?:-?\d){12,15})        # 502012-123-123-123-12
    
                 # China Union Pay CUP 19 digits
                 |62\d{4}[- ]\d{13}(?s:(?<!${ccr_csb}.{20})) # 621234-1234567890123
    
                 # ScotiBank has a break within first four digits, 16 digits total
                 |453(?s:(?<!${ccr_csb}.{3}))([- ])6\d{3}(?:\2\d{3}){3} # capture group-2, 453 6123 123 123 123, 453-6123-123-123-123
    
                 # American Express
                 |3[47]\d{2}(?:([- ])\d{4}(?s:(?<!${ccr_csb}.{9}))(?:\d\d\3\d\d  # capture group-3, 3412-123412-12123
                                                                    |\3\d{4}\3   #                  3412-1234-1234-123
                                                                  )\d{3}
                              |\d{11}(?s:(?<!${ccr_csb}.{15})))                  # 341212345678901
    
                 # JCB 15 digit flexible format
                 |(?:1800|2131)(?s:(?<!${ccr_csb}.{4}))(?:(?:${ccr_sp}?\d){11} # 1800 12345 12345 1
                                                         |(?:-?\d){11})
    
                 # Diners Club 14 digits
                 |(?:30[0-5]\d|3095|3[689]\d\d)(?:([- ])\d{6}\4\d{4}(?s:(?<!${ccr_csb}.{16}))  # capture group-4, 3001-123456-1234
                                                 |\d{10}(?s:(?<!${ccr_csb}.{14})))             #                  36121234567890
    
                 # Brazilian bank Itau Unibanco "hipercard"
                 |3841(?:([ -])\d{4}\5\d{2}\5\d{3}(?s:(?<!${ccr_csb}.{16})) # capture group-5, 3841-1234-12-123
                        |\d{9}(?s:(?<!${ccr_csb}.{13})))                    #                  3841123456789
               ) # possible CC formats ends
            ) # luhn captuer group
            ${ccr_suffix}
         |  %B\d{13,19}\^ # magnetic track format eg %B340000000000009^
       )
    /x;
    
    if (${$str} =~ /$ccr_main/) {
        my $cc_num = $1;
        unless($cc_num) { # not captured, must be in magnetic track format
            return 1; 
        }

        # Luhn check - remove spaces and dashes first
        $cc_num =~ s/[ -]//g;

        my $weight;
        my $sum = 0;
        for (my $i = 0; $i < length($cc_num) - 1; $i++) {
            $weight = substr($cc_num, -1 * ($i + 2), 1) * (2 - ($i % 2));
            $sum += (($weight < 10) ? $weight : ($weight - 9));
        }
        if (substr($cc_num, -1) == (10 - $sum % 10) % 10) {
            return 1;
        }
    }
    return 0;
}

sub expand_templates {
    my ($msg, $tmpl, $part_id) = @_;

    use PureMessage::Template qw(expand_template);
    $tmpl = expand_template(
        $tmpl,
        vars => {
                header_subject	=> scalar $msg->header('subject'),
                qid             => scalar $msg->minfo_value('Queue-Id'),
        },
    );
    return $tmpl;
}

1;

